

typedef float RealType;

struct point3d
{
  RealType x,y,z;
  point3d() {};
  point3d(RealType x, RealType y, RealType z)
  {
    this->x=x; this->y=y; this->z=z;
  };
  point3d &Normalize() {
    float t=1/sqrt(x*x+y*y+z*z);
    x*=t;
    y*=t;
    z*=t;
    return *this;
  }
};


class Matrix
{
  //friend Matrix operator* (const Matrix &a, const Matrix &b);
  //friend point3d operator* (const Matrix &a, const point3d &b);
  public:
    RealType m[4][4];
    Matrix (float dia=1);
    static Matrix Translate(RealType x, RealType y, RealType z);
    static Matrix Rotatex(RealType a);
    static Matrix Rotatey(RealType a);
    static Matrix Rotatez(RealType a);
    void LoadToGL();
    void MultGL();
};


//-----------------------------------------------------------------------------

RealType mtNorm (point3d *p)
{
  return sqrt(p->x*p->x+p->y*p->y+p->z*p->z);
  /*asm {
    //return value in st(0)
    push esi
    mov esi,p
    fld  RealTypePtr [esi+SizeOfRealType*0]
    fmul st,st(0)
    fld  RealTypePtr [esi+SizeOfRealType*1]
    fmul st,st(0)
    fadd
    fld  RealTypePtr [esi+SizeOfRealType*2]
    fmul st,st(0)
    fadd
    fsqrt
    pop esi
  }/**/
}


inline RealType mtSqrNorm (point3d *p)
{
  return (p->x*p->x+p->y*p->y+p->z*p->z);
  /*asm {
    //return value in st(0)
    push esi
    mov esi,p
    fld  RealTypePtr [esi+SizeOfRealType*0]
    fmul st,st(0)
    fld  RealTypePtr [esi+SizeOfRealType*1]
    fmul st,st(0)
    fadd
    fld  RealTypePtr [esi+SizeOfRealType*2]
    fmul st,st(0)
    fadd
    pop esi
  }/**/
}

inline point3d &operator-=(point3d &p2, const point3d &p1)
{
  p2.x-=p1.x;
  p2.y-=p1.y;
  p2.z-=p1.z;
  return p2;
}


inline point3d operator-(const point3d &p2, const point3d &p1)
{
  point3d r;
  r.x=p2.x-p1.x;
  r.y=p2.y-p1.y;
  r.z=p2.z-p1.z;
  return r;
}



inline void mtSubR(point3d *Result, const point3d *p)
{
  Result->x=p->x-Result->x;
  Result->y=p->y-Result->y;
  Result->z=p->z-Result->z;
}


inline point3d &operator+=(point3d &p1,const point3d &p2)
{
  p1.x+=p2.x;
  p1.y+=p2.y;
  p1.z+=p2.z;
  return p1;
}


inline point3d operator+(const point3d &p1,const point3d &p2)
{
  point3d r;
  r.x=p1.x+p2.x;
  r.y=p1.y+p2.y;
  r.z=p1.z+p2.z;
  return r;
}


inline point3d operator^(const point3d &p1, const point3d &p2)
{
  point3d r;
  r.x=+p1.y*p2.z-p2.y*p1.z;
  r.y=-p1.x*p2.z+p2.x*p1.z;
  r.z=+p1.x*p2.y-p2.x*p1.y;
  return r;
}


inline RealType operator*(const point3d &p1, const point3d &p2)
{
  return p1.x*p2.x+p1.y*p2.y+p1.z*p2.z;
}


//inline void mtRealMul(point3d *Result, const point3d *p1, float a)
inline point3d &operator*=(point3d &p1, RealType a)
{
  p1.x*=a;
  p1.y*=a;
  p1.z*=a;
  return p1;
}


//-----------------------------------------------------------------------------

Matrix::Matrix (float dia)
{
  for (int i=0; i<4;i++)
    for (int j=0; j<4;j++)
      m[i][j]=0;
  for (int i=0; i<4;i++)
    m[i][i]=dia;
}


Matrix Matrix::Rotatex(RealType a)
{
  Matrix x;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      x.m[i][j]=0;
  x.m[0][0]=1;
  x.m[2][2]=(RealType)cos(a);
  x.m[2][1]=-(RealType)sin(a);
  x.m[1][2]=(RealType)sin(a);
  x.m[1][1]=(RealType)cos(a);
  x.m[3][3]=1;
  return x;
}


Matrix Matrix::Rotatey(RealType a)
{
  Matrix x;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      x.m[i][j]=0;
  x.m[1][1]=1;
  x.m[0][0]=(RealType)cos(a);
  x.m[0][2]=-(RealType)sin(a);
  x.m[2][0]=(RealType)sin(a);
  x.m[2][2]=(RealType)cos(a);
  x.m[3][3]=1;
  return x;
}


Matrix Matrix::Rotatez(RealType a)
{
  Matrix x;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      x.m[i][j]=0;
  x.m[2][2]=1;
  x.m[0][0]=(RealType)cos(a);
  x.m[0][1]=-(RealType)sin(a);
  x.m[1][0]=(RealType)sin(a);
  x.m[1][1]=(RealType)cos(a);
  x.m[3][3]=1;
  return x;
}


Matrix Matrix::Translate(RealType x, RealType y, RealType z)
{
  Matrix mat;
  int i,j;
  for (i=0; i<4;i++)
    for (j=0; j<4;j++)
      mat.m[i][j]=0;
  mat.m[0][3]=x;
  mat.m[1][3]=y;
  mat.m[2][3]=z;
  mat.m[3][3]=mat.m[2][2]=mat.m[1][1]=mat.m[0][0]=1;
  return mat;
}


Matrix operator* (const Matrix &a, const Matrix &b)
{
  Matrix c;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      RealType s=0;
      for (int k=0; k<4; k++)
        s+=a.m[i][k]*b.m[k][j];
      c.m[i][j]=s;
    }
  return c;
}


point3d operator* (const Matrix &a, const point3d &b)
{
  point3d p;
  p.x=a.m[0][0]*b.x+a.m[0][1]*b.y+a.m[0][2]*b.z+a.m[0][3];
  p.y=a.m[1][0]*b.x+a.m[1][1]*b.y+a.m[1][2]*b.z+a.m[1][3];
  p.z=a.m[2][0]*b.x+a.m[2][1]*b.y+a.m[2][2]*b.z+a.m[2][3];
  return p;
}


void Matrix::LoadToGL()
{
  Matrix tmp;
  for (int i=0; i<4;i++)
    for (int j=0; j<4;j++)
      tmp.m[i][j]=this->m[j][i];
  glLoadMatrixf(tmp.m[0]);
}


void Matrix::MultGL()
{
  Matrix tmp;
  for (int i=0; i<4;i++)
    for (int j=0; j<4;j++)
      tmp.m[i][j]=this->m[j][i];
  glMultMatrixf(tmp.m[0]);
}

