

//coded by

#include <windows.h>
#include <GL/glut.h>
//#include "glut.h"
#include <vector>
#include <map>
#include <iostream>
#include <math.h>
#include <cstdio>
#pragma hdrstop

#include "mova.cpp"
#include "bmpload.cpp"


//------------------------------------------------------------------------------

class Position
{
  public:
    Matrix pos, inv_pos;
    point3d p;
    void SetE();
    void Translate(float x, float y, float z);
    void Rotatex(float angle);
    void Rotatey(float angle);
    void Rotatez(float angle);
};


void Position::SetE() { pos=1; inv_pos=1; p=point3d(0,0,0); };


void Position::Translate(float x, float y, float z)
{
  pos=pos*Matrix::Translate(x,y,z);
  inv_pos=Matrix::Translate(-x, -y, -z)*inv_pos;
}


void Position::Rotatex(float angle)
{
  pos=pos*Matrix::Rotatex(angle);
  inv_pos=Matrix::Rotatex(-angle)*inv_pos;
}


void Position::Rotatey(float angle)
{
  pos=pos*Matrix::Rotatey(angle);
  inv_pos=Matrix::Rotatey(-angle)*inv_pos;
}


void Position::Rotatez(float angle)
{
  pos=pos*Matrix::Rotatez(angle);
  inv_pos=Matrix::Rotatez(-angle)*inv_pos;
}


class Model
{
  protected:
    Texture tex;
    std::vector<point3d> points;
    std::vector<point3d> normals;
    std::vector<point3d> texps;
    struct Triangle
    {
      union {
        struct {int e0, e1, e2;};
        int e[3];
      };
      union {
        struct {int p0, p1, p2;};
        int p[3];
      };
      union {
        struct {int n0, n1, n2;};
        int n[3];
      };
      union {
        struct {int t0, t1, t2;};
        int t[3];
      };
    };
    std::vector<Triangle> trs;
  public:
    Position pos;
    virtual void SetState(double time) { };
    virtual void Draw();
};


void Model::Draw()
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  pos.pos.MultGL();
  glBindTexture(GL_TEXTURE_2D, tex.texid);
  glBegin(GL_TRIANGLES);
  for (
    std::vector<Triangle>::iterator tr_i=trs.begin();
    tr_i<trs.end();
    tr_i++
  ) {
    glTexCoord2fv(&texps[tr_i->t0].x);
    glNormal3fv(&normals[tr_i->n0].x);
    glVertex3fv(&points[tr_i->p0].x);

    glTexCoord2fv(&texps[tr_i->t1].x);
    glNormal3fv(&normals[tr_i->n1].x);
    glVertex3fv(&points[tr_i->p1].x);

    glTexCoord2fv(&texps[tr_i->t2].x);
    glNormal3fv(&normals[tr_i->n2].x);
    glVertex3fv(&points[tr_i->p2].x);
  }
  glEnd();
  glPopMatrix();
}


//-----------------------------------------------------------------------------


struct Light
{
  Position pos;
  point3d col;
  point3d point()
  {
    return point3d(
      pos.pos.m[0][3],
      pos.pos.m[1][3],
      pos.pos.m[2][3]
    );
  };
};

struct Camera
{
  Position pos;
};

class ShadeModel:
  public Model
{
  protected:
    struct Edge{
      union
      {
        struct { int a, b; };
        int p[2];
      };
      int flag;
    };
    struct Edge2i
    {
      int a, b;
      bool operator()(const Edge2i &ed1, const  Edge2i &ed2) const
      {
        return (ed1.a==ed2.a) ? (ed1.b<ed2.b) : (ed1.a<ed2.a);
      };
    };
    std::vector<Edge> edges;
    void InitEdges();
  public:
    void GenShadow(point3d light);
};


void ShadeModel::InitEdges()
{
  std::map<Edge2i, int, Edge2i> emap;
  int ed_num=0;
  for (
    std::vector<Triangle>::iterator tr_i=trs.begin();
    tr_i<trs.end();
    tr_i++
  ) {
    for (int i=0; i<3; i++) {
      Edge2i ed={tr_i->p[i], tr_i->p[(i+1)%3]};
      int t;
      if (ed.a>ed.b) {
        t=ed.a;
        ed.a=ed.b;
        ed.b=t;
      }
      std::map<Edge2i, int, Edge2i>::iterator em_i = emap.find(ed);
      if (em_i == emap.end()) {
        emap.insert(std::pair<Edge2i, int>(ed, ed_num));
        tr_i->e[i]=ed_num;
        ed_num++;
      } else {
        tr_i->e[i]=em_i->second;
      }
    }
  }
  edges.resize(ed_num);
  for (
    std::map<Edge2i, int, Edge2i>::iterator em_i=emap.begin();
    em_i!=emap.end();
    em_i++
  ) {
    edges[em_i->second].a=em_i->first.a;
    edges[em_i->second].b=em_i->first.b;
  }
}


void ShadeModel::GenShadow(point3d light)
{
  using std::vector;
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  pos.pos.MultGL();
  for (
    vector<Edge>::iterator edge_i=edges.begin();
    edge_i<edges.end();
    edge_i++
  ) {
    edge_i->flag=0;
  }
  point3d loc_light;
  loc_light=pos.inv_pos*light;
  {
    for (
      vector<Triangle>::iterator tr_i=trs.begin();
      tr_i<trs.end();
      tr_i++
    ) {
      point3d l=loc_light;
      point3d v1=points[tr_i->p1];
      point3d v2=points[tr_i->p2];
      point3d v0=points[tr_i->p0];
      v1-=v0;
      v2-=v0;
      l-=v0;
      if ((v1^v2)*l > 0) {
        for (int i=0; i<3; i++) {
          edges[tr_i->e[i]].flag++;
          edges[tr_i->e[i]].a=tr_i->p[i];
          edges[tr_i->e[i]].b=tr_i->p[(i+1)%3];
        }
        glColor3f(0,0,0);
      }
    }
  }
  {
    for (int face=0; face!=2; face++) {
      if (0==face) {
        glCullFace(GL_BACK);
        glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
      } else {
        glCullFace(GL_FRONT);
        glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
      }
      glBegin(GL_QUADS);
      for (
        vector<Edge>::iterator edge_i=edges.begin();
        edge_i<edges.end();
        edge_i++
      ) {
        if (0!=(edge_i->flag & 1)) { //not  0 or 2
          point3d va=loc_light+1000*(points[edge_i->a]-loc_light).Normalize();
          point3d vb=loc_light+1000*(points[edge_i->b]-loc_light).Normalize();
          glColor3f(0,1,1);
          glVertex3fv(
            &points[edge_i->b].x
          );
          glVertex3fv(
            &points[edge_i->a].x
          );
          glVertex3fv(
            &va.x
          );
          glVertex3fv(
            &vb.x
          );
        }
      }
      glEnd();
    }
  }
  glPopMatrix();
}

//-----------------------------------------------------------------------------

class Kolbaska:
  public ShadeModel
{
  protected:
    int m,n;
    virtual void Draw();
    void GenNormals(bool Normalize=true);
    void InitTex(int it=1, int jt=1);
    void InitTrs();
};


void Kolbaska::Draw()
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  pos.pos.MultGL();
  glBindTexture(GL_TEXTURE_2D, tex.texid);
  for (int i=0; i<m; i++) {
    glBegin(GL_TRIANGLE_STRIP);
    int ip=(i+1) % m;
    int j=0;
    int adr0=i*n;
    int adr1=ip*n;
    for (; j<n; j++) {
      glTexCoord2fv(&texps[(i+1)*(n+1)+j].x);
      glNormal3fv(&normals[adr1].x);
      glVertex3fv(&points[adr1].x);

      glTexCoord2fv(&texps[adr0+i].x);
      glNormal3fv(&normals[adr0].x);
      glVertex3fv(&points[adr0].x);
      adr0++;
      adr1++;
    }
    j=0;
    glTexCoord2fv(&texps[(i+1)*(n+1)+n].x);
    glNormal3fv(&normals[ip*n+j].x);
    glVertex3fv(&points[ip*n+j].x);

    glTexCoord2fv(&texps[i*(n+1)+n].x);
    glNormal3fv(&normals[i*n+j].x);
    glVertex3fv(&points[i*n+j].x);
    glEnd();
  }
  glPopMatrix();
}


void Kolbaska::GenNormals(bool Normalize)
{
  for (int i=0; i<m; i++) {
    int im=(i+m-1) % m;
    int ip=(i+1) % m;
    int jm=n-1;
    int jp=1;
    for (int j=0; j<n; j++) {
      normals[i*n+j]=
        (points[i*n+jp]-points[i*n+jm]) ^
        (points[ip*n+j]-points[im*n+j]);
      jm=j;
      if (++jp >= n) jp=0;
    }
  }
  if (Normalize) {
    for (int i=0; i<m*n; i++) {
      normals[i].Normalize();
    }
  }
}


void Kolbaska::InitTex(int it, int jt)
{
  texps.resize((n+1)*(m+1));
  for (int i=0; i<=m; i++) {
    for (int j=0; j<=n; j++) {
      texps[i*(n+1)+j].x=float(it)*float(i)/float(m);
      texps[i*(n+1)+j].y=float(jt)*float(j)/float(n);
    }
  }
}


void Kolbaska::InitTrs()
{
  trs.resize(2*m*n);
  for (int i=0; i<m; i++) {
    for (int j=0; j<n; j++) {
      trs[i*n+j].p0=i*n+j;
      trs[i*n+j].p1=i*n+(j+1)%n;
      trs[i*n+j].p2=((i+1)%m)*n+j;
    }
  }
  for (int i=0; i<m; i++) {
    for (int j=0; j<n; j++) {
      trs[(i+m)*n+j].p0=((i+1)%m)*n+(j+1)%n;
      trs[(i+m)*n+j].p1=((i+1)%m)*n+j;
      trs[(i+m)*n+j].p2=i*n+(j+1)%n;
    }
  }
  for (int i=0; i<2*m*n; i++) {
    trs[i].n0=trs[i].p0;
    trs[i].n1=trs[i].p1;
    trs[i].n2=trs[i].p2;
  }
  for (int i=0; i<m; i++) {
    for (int j=0; j<n; j++) {
      trs[i*n+j].t0=i*(n+1)+j;
      trs[i*n+j].t1=i*(n+1)+(j+1);
      trs[i*n+j].t2=(i+1)*(n+1)+(j);

      trs[(i+m)*n+j].t0=(i+1)*(n+1)+(j+1);
      trs[(i+m)*n+j].t1=(i+1)*(n+1)+(j);
      trs[(i+m)*n+j].t2=i*(n+1)+(j+1);
    }
  }

}




//-----------------------------------------------------------------------------

class CentrBublic:
  public Kolbaska
{
    void FillVerts(double time);
  public:
    void Init();
    virtual void SetState(double time);
};


void CentrBublic::FillVerts(double time)
{
  float rr=6.5, r=(sin(time)+1.5)*0.7;
  const float pi2= 3.14159*2;
  for (int i=0; i<m; i++) {
    float teta=pi2*i/m;
    float phi=0;
    float dphi=pi2/n;
    float cos_teta=cos(teta);
    float sin_teta=sin(teta);
    float pup=3*(sin(time*0.831)+1.1);
    float ad=i*n;
    float sin_dphi=sin(dphi);
    float cos_dphi=cos(dphi);
    float sin_phi=0;
    float cos_phi=1;
    float t;
    float sin_phi_pup_04=0;//sin(phi*pup)*0.4;
    float cos_phi_pup_04=0.4;
    float sin_dphi_pup_04=sin(dphi*pup);//sin(phi*pup)*0.4;
    float cos_dphi_pup_04=cos(dphi*pup);

    for (int j=0; j<n; j++) {
      float r2=r*(sin_phi_pup_04+1);
      points[ad].x=cos_phi*(rr+r2*cos_teta);
      points[ad].y=sin_phi*(rr+r2*cos_teta);
      points[ad].z=r2*sin_teta;
      ad++;
      phi+=dphi;
      t=cos_phi;
      cos_phi=cos_phi*cos_dphi-sin_phi*sin_dphi;
      sin_phi=sin_phi*cos_dphi+t*sin_dphi;
      t=cos_phi_pup_04;
      cos_phi_pup_04=cos_phi_pup_04*cos_dphi_pup_04-sin_phi_pup_04*sin_dphi_pup_04;
      sin_phi_pup_04=sin_phi_pup_04*cos_dphi_pup_04+t*sin_dphi_pup_04;
    }
  }
}

void CentrBublic::Init()
{
  m=20, n=40;
  Kolbaska::InitTrs();
    points.resize(m*n);
    normals.resize(n*m);
    //CentrBublic::FillVerts();
    //Kolbaska::GenNormals();
  Kolbaska::InitEdges();
  Kolbaska::InitTex(4, 16);
  tex.Load("CentrBub.bmp");
}


void CentrBublic::SetState(double time)
{
  FillVerts(time);
  GenNormals();
}

//-----------------------------------------------------------------------------
class CentrKnot
  :public Kolbaska
{
    void FillVerts();
  public:
    void Init();
};


void CentrKnot::FillVerts()
{
  float p=3;
  float q=2;
  float rr=2.4, rb=1.7, rk=0.5;
  const float pi2=3.14159*2;
  for (int i=0; i<m; i++) {
    for (int j=0; j<n; j++) {
      float phi=pi2*j/n;
      float teta=pi2*i/m;
      points[i*n+j]=
        Matrix::Rotatey(q*phi)*
        Matrix::Rotatex(0.50)*
        Matrix::Translate(rr,0,0)*
        Matrix::Rotatez(p*phi)*
        Matrix::Translate(rb,0,0)*
        Matrix::Rotatez(-teta)*
        Matrix::Translate(rk,0,0)*
        point3d(0,0,0);
      /*points[i*n+j].x=cos(phi)*(rr+r*cos(teta));
      points[i*n+j].y=sin(phi)*(rr+r*cos(teta));
      points[i*n+j].z=r*sin(teta);*/
    }
  }
}


void CentrKnot::Init()
{
  m=20, n=40;
  Kolbaska::InitTrs();
    points.resize(m*n);
    normals.resize(n*m);
    CentrKnot::FillVerts();
    Kolbaska::GenNormals();
  Kolbaska::InitEdges();
  Kolbaska::InitTex(4, 16);
  tex.Load("CentrKnot.bmp");
}

//-----------------------------------------------------------------------------

class Sobaki:
  public Kolbaska
{
    void FillVerts();
    double time;
  public:
    void Init();
    virtual void Draw();
    virtual void SetState(double time) { this->time = time; };
};


void Sobaki::FillVerts()
{
  float rr=20, r=12;
  const float pi2= 3.14159*2;
  for (int i=0; i<m; i++) {
    for (int j=0; j<n; j++) {
      float phi=pi2*j/n;
      float teta=pi2*i/m;
      points[i*n+j].x=cos(phi)*(rr+r*cos(teta));
      points[i*n+j].y=sin(phi)*(rr+r*cos(teta));
      points[i*n+j].z=r*sin(teta);
    }
  }
}


void Sobaki::Init()
{
  m=10, n=20;
  Kolbaska::InitTrs();
    points.resize(m*n);
    normals.resize(n*m);
    Sobaki::FillVerts();
    Kolbaska::GenNormals();
  Kolbaska::InitEdges();
  Kolbaska::InitTex(1, 4);
  tex.Load("Sobaka.bmp", true);
}


void Sobaki::Draw()
{
  glDisable(GL_LIGHTING);
  glEnable(GL_ALPHA_TEST);
  glDisable(GL_CULL_FACE);
  glDepthMask(GL_FALSE);
  glEnable(GL_BLEND);

  glColor4f(1,0.5*(sin(time*4.1234)+1),0.5*(cos(time*4.1234)+1),0.1);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glAlphaFunc(GL_NOTEQUAL, 0);
  Kolbaska::Draw();

  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  glEnable(GL_CULL_FACE);
  glDisable(GL_ALPHA_TEST);/**/
  glEnable(GL_LIGHTING);
}


//-----------------------------------------------------------------------------

class Room
  :public Model
{
  protected:
    int n;
    float l;
    //GLuint d_list;
  public:
    void Init();
    virtual void Draw();
};


void Room::Init()
{
  n=40;
  l=50.0;
  points.resize(n*n);
  for (int i=0; i<n; i++) {
    for (int j=0; j<n; j++) {
      float x=l*j/(n-1)-l/2;
      float z=l*i/(n-1)-l/2;
      points[i*n+j].x=x*2;
      points[i*n+j].z=z*2;
      //points[i*n+j].y=0.03*(-x*x-z*z)+0.01*(x*x*x+z*z*z);
      points[i*n+j].y=16*(sin(0.2*x)+sin(0.2*z))+0.00018*(x*x*x*x+z*z*z*z)-30;
    }
  }
  normals.resize(n*n);
  for (int i=0; i<n; i++) {
    for (int j=0; j<n; j++) {
      int ip=(i+1)%n;
      int im=(i+n-1)%n;
      int jp=(j+1)%n;
      int jm=(j+n-1)%n;
      normals[i*n+j]=
        (points[ip*n+j]-points[im*n+j]) ^
        (points[i*n+jp]-points[i*n+jm]);
      normals[i*n+j].Normalize();
    }
  }
  texps.resize(n*n);
  for (int i=0; i<n; i++) {
    for (int j=0; j<n; j++) {
      texps[i*n+j].x=float(j)/n*10;
      texps[i*n+j].y=float(i)/n*10;
    }
  }
  trs.resize(2*(n-1)*(n-1));
  for (int i=0; i<n-1; i++) {
    for (int j=0; j<n-1; j++) {
      trs[2*(i*(n-1)+j)].n0=trs[2*(i*(n-1)+j)].t0=trs[2*(i*(n-1)+j)].p0=i*n+j;
      trs[2*(i*(n-1)+j)].n1=trs[2*(i*(n-1)+j)].t1=trs[2*(i*(n-1)+j)].p1=(i+1)*n+j;
      trs[2*(i*(n-1)+j)].n2=trs[2*(i*(n-1)+j)].t2=trs[2*(i*(n-1)+j)].p2=(i)*n+(j+1);
      trs[2*(i*(n-1)+j)+1].n0=trs[2*(i*(n-1)+j)+1].t0=trs[2*(i*(n-1)+j)+1].p0=(i+1)*n+(j+1);
      trs[2*(i*(n-1)+j)+1].n1=trs[2*(i*(n-1)+j)+1].t1=trs[2*(i*(n-1)+j)+1].p1=(i)*n+(j+1);
      trs[2*(i*(n-1)+j)+1].n2=trs[2*(i*(n-1)+j)+1].t2=trs[2*(i*(n-1)+j)+1].p2=(i+1)*n+j;
    }
  }
  tex.Load("floor.bmp");
}


void Room::Draw()
{
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  pos.pos.MultGL();
  glBindTexture(GL_TEXTURE_2D, tex.texid);
  for (int i=0; i<n-1; i++) {
    glBegin(GL_TRIANGLE_STRIP);
    for (int j=0; j<n; j++) {
      glTexCoord2fv(&texps[i*n+j].x);
      glNormal3fv(&normals[i*n+j].x);
      glVertex3fv(&points[i*n+j].x);

      glTexCoord2fv(&texps[(i+1)*n+j].x);
      glNormal3fv(&normals[(i+1)*n+j].x);
      glVertex3fv(&points[(i+1)*n+j].x);
    }
    glEnd();
  }
  glPopMatrix();
}


//-----------------------------------------------------------------------------
class Scene
{
    std::vector<ShadeModel *> sh_mdls;
    std::vector<Model *> mdls;
    std::vector<Light *> lights;
    CentrBublic centr_bublic;
    CentrKnot centr_knot;
    Sobaki sobaki;
    Room room;
    Camera cam;
    Light my_lights[3];
    double time;
    void DrawShadow();
    int LightMode; //0 - 3 lights; 1 - 3 lights, no shadow, 2 - one white light
  public:
    void Init();
    ~Scene();
    void SetState(double Time);
    void Render();
    void ChangeLights();
};


void Scene::Init()
{
  LightMode=2;

  centr_bublic.Init();
  room.Init();
  centr_knot.Init();
  sobaki.Init();

  mdls.push_back(&centr_bublic);
  sh_mdls.push_back(&centr_bublic);

  room.pos.SetE();
  mdls.push_back(&room);

  mdls.push_back(&centr_knot);
  sh_mdls.push_back(&centr_knot);

  mdls.push_back(&sobaki);
}


Scene::~Scene()
{
}


void Scene::SetState(double time)
{
  this->time=time;
  cam.pos.SetE();
  cam.pos.Rotatey(time*1.1);
  cam.pos.Rotatex((sin(time*0.9293)+1)*0.4+0.5);
  cam.pos.Translate(0,0,25+sin(time*0.29831275)*15);

  centr_bublic.pos.SetE();
  centr_bublic.pos.Translate(0, -1.6, 0);
  centr_bublic.pos.Rotatex(time);

  centr_knot.pos.SetE();
  centr_knot.pos.Translate(0, -1.6, 0);

  sobaki.pos.SetE();
  sobaki.pos.Translate(0,0,0);
  sobaki.pos.Rotatex(1.57);


  float t=sin(time*1.123)*0.8;
  t=t*t*t;
  float dx=sin(time*0.723)*3;
  float dz=cos(time*0.723)*3;

  lights.resize(0);

  if (LightMode!=2) {
    my_lights[0].col=point3d(0.8,0,0);
    my_lights[1].col=point3d(0,0.8,0);
    my_lights[2].col=point3d(0,0,0.8);
    my_lights[0].pos.SetE(); my_lights[0].pos.Translate(0*t+dx,6.6,-6*t+dz);
    my_lights[1].pos.SetE(); my_lights[1].pos.Translate(4*t+dx,6.6,3*t+dz);
    my_lights[2].pos.SetE(); my_lights[2].pos.Translate(-4*t+dx,6.6,3*t+dz);
    for (int i=0; i<3; i++)
      lights.push_back(my_lights+i);
  } else {
    my_lights[0].pos.SetE(); my_lights[0].pos.Translate(0*t+dx,6.6,-6*t+dz);
    my_lights[0].col=point3d(0.8,0.8,0.8);
    lights.push_back(my_lights+0);
  }
  for (
    std::vector<Model *>::iterator m_i=mdls.begin();
    m_i!=mdls.end();
    m_i++
  ) {
    (**m_i).SetState(time);
  }
}


void Scene::Render()
{
  glClearColor(0.0,0.00,0,1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  cam.pos.inv_pos.LoadToGL();
  {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    {
      glEnable(GL_TEXTURE_2D);
      glEnable(GL_LIGHTING);
      for (unsigned l_i=0; l_i<lights.size(); l_i++) {
        glEnable(GL_LIGHT0+l_i);
        point3d p=lights[l_i]->col;
        float col[4]={p.x, p.y, p.z, 1};
        glLightfv(GL_LIGHT0+l_i, GL_DIFFUSE, col );
        p=lights[l_i]->point();
        float pos[4]={p.x, p.y, p.z, 1};
        glLightfv(GL_LIGHT0+l_i, GL_POSITION, pos);
        glLightf(GL_LIGHT0+l_i, GL_CONSTANT_ATTENUATION, 1.0*0.3);
        glLightf(GL_LIGHT0+l_i, GL_LINEAR_ATTENUATION, 0.2*0.3);
      }
      for (
        std::vector<Model *>::iterator m_i=mdls.begin();
        m_i!=mdls.end();
        m_i++
      ) {
        (**m_i).Draw();
      }
      for (unsigned l_i=0; l_i<lights.size(); l_i++) {
        glDisable(GL_LIGHT0+l_i);
      }
      glDisable(GL_LIGHTING);
      glDisable(GL_TEXTURE_2D);
    }
    if (LightMode!=1)
      DrawShadow();
    {
      glPointSize(10);
      glBegin(GL_POINTS);
      for (unsigned l_i=0; l_i<lights.size(); l_i++) {
          point3d lp = lights[l_i]->point();
          glColor3fv(&lights[l_i]->col.x);
          glVertex3fv(&lp.x);
      }
      glEnd();
    }
    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
  }
}


void Scene::DrawShadow()
{
  glEnable(GL_STENCIL_TEST);
  glClear(GL_STENCIL_BUFFER_BIT);
  glDepthMask(0);
  for (
    std::vector<Light *>::iterator l_i=lights.begin();
    l_i!=lights.end();
    l_i++
  ) {
    glStencilFunc(GL_ALWAYS, 0, 0);
    glColorMask(0,0,0,0);
    for (
      std::vector<ShadeModel *>::iterator sm_i=sh_mdls.begin();
      sm_i!=sh_mdls.end();
      sm_i++
    ) {
      (**sm_i).GenShadow((*l_i)->point());
    }
    glCullFace(GL_BACK);
    {
      glEnable(GL_BLEND);
      glColorMask(1,1,1,1);
      glBlendFunc(GL_DST_COLOR, GL_ZERO);
      glColor4f(1-(*l_i)->col.x,1-(*l_i)->col.y,1-(*l_i)->col.z, 1);
      glStencilFunc(GL_NOTEQUAL, 0, ~0);
      glStencilOp(GL_KEEP, GL_ZERO, GL_KEEP); 
      glMatrixMode (GL_MODELVIEW);
      glPushMatrix ();
        glLoadIdentity ();
        glMatrixMode (GL_PROJECTION);
        glPushMatrix ();
          glLoadIdentity ();
          glBegin (GL_QUADS);
            glVertex3i (-1, -1, -1);
            glVertex3i (1, -1, -1);
            glVertex3i (1, 1, -1);
            glVertex3i (-1, 1, -1);
          glEnd ();
        glPopMatrix ();
        glMatrixMode (GL_MODELVIEW);
      glPopMatrix ();
      glDisable(GL_BLEND);
    }
  }
  glDepthMask(1);
  glDisable(GL_STENCIL_TEST);
}


void Scene::ChangeLights()
{
  LightMode=(LightMode+1)%3;
}


//---------------------------------------------------------------------------


Scene scene;
unsigned long start_time;
int frame_count;

void Key(unsigned char key, int, int)
{
  switch (key) {
    case 's':
    case 'S':
      scene.ChangeLights();
      break;
    case 27: {
      /*char buf[20];
      double fps=1000.0 * frame_count / ((GetTickCount()-start_time));
      std::sprintf(buf, "%f", fps);
      MessageBox(0, buf, "FPS:", MB_OK);*/
      exit(0);
    }
  }
}


void OnIdle()
{
  glutPostRedisplay();
}


void Reshape(int w, int h)
{
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  float aspect=float(w)/h;
  gluPerspective(90/aspect, aspect, 1, 100);
}


void Display()
{
  double time = (GetTickCount()-start_time)/1000.0;
  scene.SetState(time);
  scene.Render();
  glutSwapBuffers();
  frame_count++;
}


int main(int argc, char* argv[])
{
  std::cout << "hello\n";
  std::vector<point3d> vi;
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL);
  glutInitWindowSize(800, 600);
  //glutInitWindowSize(1280, 1024);
  glutCreateWindow("Shadow Knot");
  //glutFullScreen();

  glutDisplayFunc(Display);
  glutReshapeFunc(Reshape);
  glutKeyboardFunc(Key);
  glutIdleFunc(OnIdle);

  start_time=GetTickCount();
  int  s_bits;
  glGetIntegerv(GL_STENCIL_BITS, &s_bits);
  if (s_bits<4) {
    MessageBox(0, "Stencil buffer bits is less than 4. There will be not accurate shadows.", "Warning", MB_OK);
  }
  try {
    scene.Init();
    frame_count=0;
    glutMainLoop();
  } catch (char * mess) {
    std::cerr << mess;
  }
  return 0;
}               


/**/